package android.image;

import android.world.Posn;
import android.graphics.Canvas;
import android.graphics.Paint;

/** A abstract Class representing an Image.  See subclasses for
 *    specific implementations. */
public abstract class Image{
    protected static int OUTLINE = 0;
    protected static int SOLID = 1;
    
    /** Pinhole for alignment to other images */
    protected int pinholeX, pinholeY;
    /** Only for subclasses */
    protected Image(int px, int py){
        pinholeX = px;
        pinholeY = py;
    }
    // Calculate Image size to the left/right/up/down of the pinhole
    protected int leftOfPin(){ return pinholeX; }
    protected int rightOfPin(){ return width()-pinholeX; }
    protected int upOfPin(){ return pinholeY; }
    protected int downOfPin(){ return height()-pinholeY; }
    
    /** Overlay the given Image over this Image */
    public Image overlay(Image top)
    { return new Overlay(top,this); }
    /** Overlay the given Images over this Image */
    public Image overlay(Image top, Image next, Image ... imgs){
        return new Overlay(new Overlay(top, make(next, imgs, 0)), this);
    }
    /** Overlay the given Image over this Image offset (x,y) */
    public Image overlayxy(Image top, int x, int y)
    { return new OverlayXY(top,x,y,this); }
    /** Overlay the given Image over this Image offset (x,y) */
    public Image overlayxy(Image top, double x, double y)
    { return overlayxy(top,(int)x,(int)y); }
    /** Overlay the given Image over this Image offset Posn */
    public Image overlayxy(Image top, Posn p)
    { return overlayxy(top,p.x,p.y); }

    /** Overlay the given Image over this Image offset (x,y) */
    public Image overlayxy(int x, int y, Image top)
    { return overlayxy(top,x,y); }
    /** Overlay the given Image over this Image offset (x,y) */
    public Image overlayxy(double x, double y, Image top)
    { return overlayxy(top,(int)x,(int)y); }
    /** Overlay the given Image over this Image offset Posn */
    public Image overlayxy(Posn p, Image top)
    { return overlayxy(top,p.x,p.y); }
    
    /** Make a layered Image with top t and bottoms of imgs[i..] */
    private static Image make(Image t, Image[] imgs, int i){
        if(i == imgs.length)
            return t;
        return new Overlay(t, make(imgs[i], imgs, i+1));    
    }
    
    
    /** Draw this image into a Graphics */
    public abstract void paint(Canvas c, int x, int y);
    /** Calculate the width of this Image */
    public abstract int width();
    /** Calculate the height of this Image */
    public abstract int height();
    
    /** Convert this image into a Scene containing the image */
    public Scene toScene(){
        return new MT(this.width()+2, this.height()+2).placeImage(this, pinholeX+1, pinholeY+1);
    }
    /** Represents a truly empty scene... */
    private static class MT extends EmptyScene{
        MT(int w, int h){ super(w,h); }
    }
    
    /** Save this image to a file */
    public boolean toFile(String name){
        return toScene().toFile(name);
    }
    
    /** Convert the mode string into an int representing Outline/Solid */
    protected static int mode(String s){
        if(s.equals("outline")) return OUTLINE;
        if(s.equals("solid")) return SOLID;
        throw new RuntimeException("Mode expected \"solid\" or \"outline\", got: \""+s+"\"");
    }

    /** Round the given Double to an Int */
    protected static int round(double d){
        return (int)Math.round(d);
    }

    /** Get the java.awt.Color represented by the given string */
    protected static int color(String s){
        return ColorDatabase.color(s);
    }
    
    /** Create a Paint */
    protected static Paint painter(int color, int style){
        Paint p = new Paint();
        p.setColor(color);
        p.setAntiAlias(true);
        p.setStyle(style == Image.OUTLINE?Paint.Style.STROKE:Paint.Style.FILL);
        return p;
    }
}
