package android.world;

import android.app.Activity;
import android.image.Scene;

/**
 * <style type='text/css'><!--
 *    .def{ color: #000000; }
 *    .com{ font-style: italic; color: #880000; }
 *    .keyw{ font-weight: bold; color: #000088; }
 *    .num{ color: #00AA00; }
 *    .str{ color: #CC00AB; }
 *    .prim{ color: #0000FF; }
 *    .func{ color: #BB7733; }
 *    img.example{ padding-left: 50px; padding-bottom: 30px; }
 *    table.events td{ padding-bottom: 20px; }
 *    table.events{ padding-left: 30px; padding-bottom: 20px; }
 *  --></style>
 *  
 *  A Class representing an imperative World and the related methods for drawing the world
 *    and handling various events.  In order to implement a functioning World you
 *    must <i>extend</i> this class, and implement an {@link android.world.VoidWorld#onDraw onDraw}
 *    method.  Other handler methods ({@link android.world.VoidWorld#tickRate tickRate}, {@link android.world.VoidWorld#onTick onTick}, 
 *    {@link android.world.VoidWorld#onMouse onMouse}, {@link android.world.VoidWorld#onKey onKey}, {@link android.world.VoidWorld#onRelease onRelease}, 
 *    {@link android.world.VoidWorld#onOrientation onOrientation}, {@link android.world.VoidWorld#stopWhen stopWhen},
 *    and {@link android.world.VoidWorld#lastScene lastScene}) are optional, and can be overridden to add new functionality.
 * <p>
 *  See the individual methods for detailed documentation.
 * </p>
 * 
 * <p>
 * <h3>Extending VoidWorld</h3>
 * <blockquote>
 * Below is a simple example of a <code>World</code> that adds a new point at each mouse click.  The world
 * contains a {@link android.image.Scene Scene} and a new {@link android.image.Circle Circle} is placed for each
 * <code class='str'>"button-down"</code> event received. 
 * 
 * 
 * <pre>   
 *        <span class="keyw">import</span> android.app.Activity;
 *        <span class="keyw">import</span> android.os.Bundle;
 *       
 *        <span class="keyw">import</span> android.image.*;
 *        <span class="keyw">import</span> android.world.World;
 *       
 *        <span class="keyw">public</span> <span class="keyw">class</span> MousePoints <span class="keyw">extends</span> Activity{
 *            <span class="com">// Called when the activity is first created.</span>
 *            <span class="keyw">public</span> <span class="keyw">void</span> <span class="func">onCreate</span>(Bundle savedState) {
 *                <span class="keyw">super</span>.<span class="func">onCreate</span>(savedState);
 *               
 *                <span class="com">// Create and start the World </span>
 *                <span class="keyw">new</span> <span class="func">MousePointsWorld</span>(<span class="keyw">new</span> <span class="func">EmptyScene</span>(<span class="num">320</span>, <span class="num">480</span>))
 *                       .<span class="func">bigBang</span>(<span class="keyw">this</span>);        
 *            }
 *        }
 *       
 *        <span class="keyw">public</span> <span class="keyw">class</span> MousePointsVoidWorld <span class="keyw">extends</span> VoidWorld{
 *            <span class="keyw">public</span> <span class="keyw">static</span> <span class="keyw">void</span> <span class="func">main</span>(<span class="prim">String</span>[] args){
 *                <span class="keyw">new</span> <span class="func">MousePointsVoidWorld</span>().<span class="func">bigBang</span>();
 *            }
 *            
 *            Scene scene = <span class="keyw">new</span> <span class="func">EmptyScene</span>(<span class="num">200</span>, <span class="num">200</span>);
 *     
 *            <span class="keyw">public</span> Scene <span class="func">onDraw</span>(){ <span class="keyw">return</span> <span class="keyw">this</span>.scene; }
 *     
 *            <span class="keyw">public void</span> <span class="func">onMouse</span>(<span class="keyw">int</span> x, <span class="keyw">int</span> y, <span class="prim">String</span> me){
 *                <span class="keyw">if</span>(me.<span class="func">equals</span>(<span class="str">"button-down"</span>)){
 *                    <span class="keyw">this</span>.scene = <span class="keyw">this</span>.scene.<span class="func">placeImage</span>(
 *                                     <span class="keyw">new</span> <span class="func">Circle</span>(<span class="num">20</span>, <span class="str">"solid"</span>, <span class="str">"red"</span>)
 *                                           .<span class="func">overlay</span>(<span class="keyw">new</span> <span class="func">Circle</span>(<span class="num">20</span>, <span class="str">"outline"</span>, <span class="str">"black"</span>)), x, y);
 *                }
 *            }
 *        }
 * </pre>
 * 
 * After a few finger-taps, the device will look something like this:<br/><br/>
 * 
 *   <img class="example" src="test/void-mouse-points.png" />
 * </blockquote>
 * </p>    
 */
public abstract class VoidWorld{
    
    /** Default Tick rate for the world: ~33 frames per second */
    public static double DEFAULT_TICK_RATE = 0.03;

    /** Mouse down (button-down) event String */
    public static String MOUSE_DOWN = BigBang.MOUSE_DOWN;
    /** Mouse up (button-up) event String */
    public static String MOUSE_UP = BigBang.MOUSE_UP;
    /** Mouse motion (move) event String */
    public static String MOUSE_MOVE = BigBang.MOUSE_MOVE;
    /** Mouse down & move (drag) event String */
    public static String MOUSE_DRAG = BigBang.MOUSE_DRAG;   

    /** Key arrow-up event String */
    public static String KEY_ARROW_UP = BigBang.KEY_ARROW_UP;
    /** Key arrow-down event String */
    public static String KEY_ARROW_DOWN = BigBang.KEY_ARROW_RIGHT;
    /** Key arrow-left event String */
    public static String KEY_ARROW_LEFT = BigBang.KEY_ARROW_LEFT;
    /** Key arrow-right event String */
    public static String KEY_ARROW_RIGHT = BigBang.KEY_ARROW_RIGHT;
    
    /** Return a visualization of this <tt>World</tt> as a {@link android.image.Scene Scene}.
     *    See {@link android.image.EmptyScene}, {@link android.image.Scene#placeImage(android.image.Image, int, int)}, and
     *    {@link android.image.Scene#addLine(int, int, int, int, String)} for documentation on
     *    constructing <tt>Scene</tt>s */
    public abstract Scene onDraw();
    
    /** Return the tick rate for this World in <i>seconds</i>.  For example,
     *  <span class='num'>0.5</span> means two <i>tick</i>. The rate is only accessed when
     *  bigBang() is initially called and the window is created. */
    public double tickRate(){ return DEFAULT_TICK_RATE; }
    
    /** Change this World based on the Tick of the clock.  This
     *  method is called to get the update the World on each clock tick.*/
    public void onTick(){ }
    
    /** Change this World when a mouse event is triggered.
     * <tt>x</tt> and <tt>y</tt> are the location of the event in the window, and 
     * <tt>event</tt> is a <tt>String</tt> that describes what kind of event
     * occurred.
     * 
     * <p>
     *   <b>Possible Mouse Events</b>
     * <table class='events'>
     *    <tr><td style="text-align:right"><tt class='str'>"button-down"</tt> : </td>
     *        <td>The user <i>presses</i> a mouse button in the World window</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"button-up"</tt> : </td>
     *        <td>The user <i>releases</i> a mouse button in the World window</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"move"</tt> : </td>
     *        <td>The user <i>moves</i> the mouse in the World window</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"drag"</tt> : </td>
     *        <td>The user <i>holds</i> a mouse button and <i>moves</i> the mouse in the World window</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"enter"</tt> : </td>
     *        <td>The user moves the mouse <i>in-to</i> the World window</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"leave"</tt> : </td>
     *        <td>The user moves the mouse <i>out-of</i> the World window</td></tr>
     * </table>
     * </p>
     */
    public void onMouse(int x, int y, String event){ }
    
    /** Change this World when a key event is
     * triggered. The given <tt>event</tt> is a <tt>String</tt> that
     * describes which key was pressed.
     *
     * <p>
     *   <b>Special Key</b>
     * <table class='events'>
     *    <tr><td style="text-align:right"><tt class='str'>"up"</tt> : </td>
     *        <td>The user presses the <i>up-arrow</i> key</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"down"</tt> : </td>
     *        <td>The user presses the <i>down-arrow</i> key</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"left"</tt> : </td>
     *        <td>The user presses the <i>left-arrow</i> key</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"right"</tt> : </td>
     *        <td>The user presses the <i>right-arrow</i> key</td></tr>
     * </table>
     *
     * Other keys generate a single character <tt>String</tt> that
     * represents the key pressed. For example, Pressing the <i>B</i> key on
     * the keyboard generates <tt class='str'>"b"</tt> as an event.
     * If the shift key is held while pressing <i>B</i> then <tt
     * class='str'>"B"</tt> is generated.
     * </p>
     */
    public void onKey(String event){ }

    /** Change this World when a key is released. The given <tt>event</tt>
     * is a <tt>String</tt> that describes which key was released.
     *
     * <p>
     *   <b>Special Keys</b>
     * <table class='events'>
     *    <tr><td style="text-align:right"><tt class='str'>"up"</tt> : </td>
     *        <td>The user presses the <i>up-arrow</i> key</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"down"</tt> : </td>
     *        <td>The user presses the <i>down-arrow</i> key</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"left"</tt> : </td>
     *        <td>The user presses the <i>left-arrow</i> key</td></tr>
     *    <tr><td style="text-align:right"><tt class='str'>"right"</tt> : </td>
     *        <td>The user presses the <i>right-arrow</i> key</td></tr>
     * </table>
     *
     * Other keys generate a single character <tt>String</tt> that
     * represents the key released. For example, Pressing then releasing the <i>B</i> key on
     * the keyboard generates <tt class='str'>"b"</tt> as an <tt>onKey</tt> event and again
     * as an <tt>onRelease</tt> event.  If the shift key is held while pressing/releasing <i>B</i> then <tt
     * class='str'>"B"</tt> is generated.
     * </p>
     */
    public void onRelease(String event){ }

    /** Change this World when a device orientation event is triggered.
     * <tt>x</tt>, <tt>y</tt>, and <tt>z</tt> make-up the new orientation vector 
     * for the device.  When the device is resting on a flat, level surface the
     * orientation should be entirely in the <tt>z</tt> direction (typically
     * straight out the the back of the device).
     * 
     * <p>
     *   If you are only concerned with the "<i>down</i>" angle (direction of down
     *   with respect to the screen) you can calculate it with a method as follows:
     *   <pre>
     *      <span class='keyw'>double</span> down(<span class='keyw'>float</span> x, <span class='keyw'>float</span> y, <span class='keyw'>float</span> z){
     *         <span class='keyw'>double</span> ang = Math.acos(x/Math.sqrt(x*x + y*y))+Math.PI;
     *         <span class='keyw'>if</span>(y &lt; <span class='num'>0</span>)<span class='keyw'>return</span> <span class='num'>2</span>*Math.PI-ang;
     *         <span class='keyw'>return</span> ang;
     *      }
     *   </pre>
     * </p>
     */
    public void onOrientation(float x, float y, float z){ }
    
    /** Determine if the World/interaction/animation should be
     * stopped.  Returning a value of <tt class='keyw'>true</tt>
     * discontinues all events (mouse, key, ticks) and causes {@link
     * android.world.VoidWorld#lastScene} to be used to draw the final
     * <tt>Scene</tt>.
     */
    public boolean stopWhen(){ return false; }
    
    /** Returns the <tt>Scene</tt> that should be displayed when the
     * interaction/animation completes ({@link android.world.VoidWorld#stopWhen}
     * returns <tt class='keyw'>true</tt>). */
    public Scene lastScene(){ return this.onDraw(); }
    
    /** Kick off the interaction/animation.  This method returns the final
     *    state of the world after the user closes the World window. */
    public VoidWorld bigBang(Activity act){
        return (VoidWorld)new BigBang(this)
            .onDraw(new WorldDraw())
            .onTick(new WorldTick(), tickRate())
            .onMouse(new WorldMouse())
            .onKey(new WorldKey())
            .stopWhen(new WorldStop())
            .lastScene(new WorldLast())
            .orientation(new WorldOrient())
            .bigBang(act);
    }
    
    /** Wrapper for OnDraw callback */
    private static class WorldDraw{
        Scene apply(VoidWorld w){ return w.onDraw(); }
    }
    /** Wrapper for OnTick callback */
    private static class WorldTick{
        VoidWorld apply(VoidWorld w){ w.onTick(); return w; }
    }
    /** Wrapper for OnMouse callback */
    private static class WorldMouse{
        VoidWorld apply(VoidWorld w, int x, int y, String me)
        { w.onMouse(x,y,me); return w; }
    }
    /** Wrapper for OnKey callback */
    private static class WorldKey{
        VoidWorld apply(VoidWorld w, String ke){ w.onKey(ke); return w; }
    }
    /** Wrapper for OnKey callback */
    private static class WorldRelease{
        VoidWorld apply(VoidWorld w, String ke){ w.onRelease(ke); return w; }
    }
    /** Wrapper for StopWhen callback */
    private static class WorldStop{
        boolean apply(VoidWorld w){ return w.stopWhen(); }
    }
    /** Wrapper for LastScene callback */
    private static class WorldLast{
        Scene apply(VoidWorld w){ return w.lastScene(); }
    }
    /** Wrapper for LastScene callback */
    private static class WorldOrient{
        VoidWorld apply(VoidWorld w, float x, float y, float z){ w.onOrientation(x,y,z); return w; }
    }
    /** Make sure that changes are redrawn every time */
    public boolean equals(Object o){ return false; }
}

